Namespace Common

    Namespace [String]

        <HideModuleName()> _
        Module [String]

            ''' <summary>
            ''' Splits a string into an array of substring on the basis of a <paramref name="separator"></paramref> string.
            ''' </summary>
            ''' <param name="str">The string to be split.</param>
            ''' <param name="separator">The separator string.</param>
            ''' <returns>Returns an array of the resulting string parts.</returns>
            Public Function Split(ByVal str As String, ByVal separator As String) As String()
                Return str.Split(New String() {separator}, StringSplitOptions.None)
            End Function

            ''' <summary>
            ''' Splits a string into an array of substrings of maximum <paramref name="cluster_size"></paramref> size.
            ''' </summary>
            ''' <param name="str">The string to be split.</param>
            ''' <param name="cluster_size">The size of a cluster</param>
            ''' <returns>Returns an array of the resulting string parts.</returns>
            Public Function Split(ByVal str As String, ByVal cluster_size As Integer) As String()
                Dim ret As New List(Of String)

                Dim pos As Integer = 0

                Do
                    If pos + cluster_size > str.Length Then
                        ret.Add(str.Substring(pos))
                        Exit Do
                    End If
                    ret.Add(str.Substring(pos, cluster_size))
                    pos += cluster_size
                Loop

                Return ret.ToArray()
            End Function

            Public Function SplitLines(ByVal str As String, ByVal length As Integer) As String()
                Dim tokens As String() = str.Split(" "c)
                Dim ret As New List(Of String)
                Dim current As New Text.StringBuilder()
                Dim i As Integer = 0

                Do Until i = tokens.Length
                    Do While current.Length + tokens(i).Length < length
                        current.Append(tokens(i))
                        current.Append(" "c)
                        i += 1

                        If i = tokens.Length Then Exit Do
                    Loop

                    If current.Length = 0 Then
                        current.Append(tokens(i))
                        i += 1
                    End If

                    ret.Add(current.ToString())
                    current.Remove(0, current.Length)
                Loop

                Return ret.ToArray()
            End Function

            ''' <summary>
            ''' Replaces all ocurrences of a certain substring by a new string.
            ''' </summary>
            ''' <param name="str">The string to search and replace in.</param>
            ''' <param name="old_value">The old substring.</param>
            ''' <param name="new_value">The new substring.</param>
            ''' <returns>Returns the string with all replaces done.</returns>
            ''' <remarks>
            ''' Unlike <see cref="String.Replace"></see>, this version re-searches
            ''' the new string after having completed the replacement to test, whether
            ''' (through the replacement operation) the old search string is again a
            ''' substring of the string, in which case the replacement cycle is repeated.
            ''' This may lead to an endless loop if <paramref name="new_value"></paramref>
            ''' contains <paramref name="old_value"></paramref> as a substring.
            ''' </remarks>
            Public Function ReplaceAll(ByVal str As String, ByVal old_value As String, ByVal new_value As String) As String
                Dim ret As String = str

                Do While ret.IndexOf(old_value) > -1
                    ret = ret.Replace(old_value, new_value)
                Loop

                Return ret
            End Function

            ''' <summary>
            ''' Converts an array of strings to an array of characters.
            ''' </summary>
            ''' <param name="strings">The array of strings.</param>
            ''' <returns>Returns an array of chars.</returns>
            ''' <remarks>
            ''' This function converts all items of the array using the conventional cast
            ''' operation for <see cref="String"></see> to <see cref="Char"></see>.
            ''' </remarks>
            Public Function ArrayToChars(ByVal strings() As String) As Char()
                Dim ret As New List(Of Char)

                For Each s As String In strings
                    ret.Add(CChar(s))
                Next

                Return ret.ToArray()
            End Function

            ''' <summary>
            ''' Reverses a string.
            ''' </summary>
            ''' <param name="str">The string.</param>
            ''' <returns>Returns the reversed string.</returns>
            Public Function Reverse(ByVal str As String) As String
                Dim ret As New Text.StringBuilder()

                For i As Integer = str.Length - 1 To 0 Step -1
                    ret.Append(str(i))
                Next

                Return ret.ToString()
            End Function

        End Module

    End Namespace

End Namespace