Namespace Calculate.Expressions

    Public Enum ExpressionFormat
        Infix
        Prefix
        Postfix
    End Enum

    Public MustInherit Class Expression

#Region " Public "

#Region " Constructors "

        <DebuggerHidden()> _
        Public Sub New(ByVal expression As String, Optional ByVal start As Integer = 0)
            Me.Expression = expression
            Me.Start = start
        End Sub

#End Region

#Region " Properties "

        Public Property Expression() As String
            <DebuggerHidden()> _
            Get
                Return _expression
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As String)
                _expression = value
            End Set
        End Property

        Public Property Start() As Integer
            <DebuggerHidden()> _
            Get
                Return _start
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As Integer)
                _start = value
            End Set
        End Property

        Public Shared Property LastError() As Exception
            <DebuggerHidden()> _
            Get
                Return _last_error
            End Get
            <DebuggerHidden()> _
            Private Set(ByVal value As Exception)
                _last_error = value
            End Set
        End Property

#End Region

#Region " Methods "

        ''' <summary>
        ''' Parses the expression.
        ''' </summary>
        ''' <exception cref="ParseException">Thrown if the expression is not valid.</exception>
        ''' <returns>Returns the parsed expression.</returns>
        <DebuggerHidden()> _
        Public Function Parse() As ParsedExpression
            Dim ret As ParsedExpression = Nothing
            If Not TryParse(ret) Then Throw [Error]
            Return ret
        End Function

        ''' <summary>
        ''' Tries to parse the expression.
        ''' </summary>
        ''' <param name="ret">Holds the parsed expression, if parsing was successful.</param>
        ''' <returns>Returns whether parsing was successful.</returns>
        Public MustOverride Function TryParse(ByRef ret As ParsedExpression) As Boolean

#Region " Static "

        ''' <summary>
        ''' Parses the <paramref name="expression"></paramref>.
        ''' </summary>
        ''' <param name="expression">The expression to parse.</param>
        ''' <param name="format">The expression's format.</param>
        ''' <returns>Returns the parsed expression.</returns>
        ''' <exception cref="ParseException">Thrown if the expression is not valid.</exception>
        <DebuggerHidden()> _
        Public Shared Function Parse( _
            ByVal expression As String, _
            Optional ByVal start As Integer = 0, _
            Optional ByVal format As ExpressionFormat = ExpressionFormat.Infix _
        ) As ParsedExpression
            Return ExpressionFromFormat(expression, start, format).Parse()
        End Function

        ''' <summary>
        ''' Tries to parse the <paramref name="expression"></paramref>.
        ''' </summary>
        ''' <param name="expression">The expression to parse.</param>
        ''' <param name="format">The expression's format.</param>
        ''' <param name="ret">The parsed expression, if successful.</param>
        ''' <returns>Returns whether parsing was successful.</returns>
        <DebuggerHidden()> _
        Public Shared Function TryParse( _
            ByVal expression As String, _
            ByVal format As ExpressionFormat, _
            ByRef ret As ParsedExpression, _
            Optional ByVal start As Integer = 0 _
        ) As Boolean
            With ExpressionFromFormat(expression, start, format)
                If Not .TryParse(ret) Then
                    LastError = .Error
                    Return False
                Else
                    Return True
                End If
            End With
        End Function

        ''' <summary>
        ''' Tries to parse the <paramref name="expression"></paramref>.
        ''' </summary>
        ''' <param name="expression">The expression to parse.</param>
        ''' <param name="ret">The parsed expression, if successful.</param>
        ''' <returns>Returns whether parsing was successful.</returns>
        ''' <remarks>Treats the expression as an infix expression.</remarks>
        <DebuggerHidden()> _
        Public Shared Function TryParse(ByVal expression As String, ByRef ret As ParsedExpression) As Boolean
            Return TryParse(expression, ExpressionFormat.Infix, ret)
        End Function

#End Region

#End Region

#End Region

#Region " Private "

        Private _expression As String
        Private _start As Integer
        Private _scanner As Parser.Scanner
        Private _stack As Stack
        Private _error As Exception

        Private Shared _last_error As Exception

#Region " Properties "

        Protected Property Stack() As Stack
            <DebuggerHidden()> _
            Get
                Return _stack
            End Get
            <DebuggerHidden()> _
            Set(ByVal value As Stack)
                _stack = value
            End Set
        End Property

        Protected Property [Error]() As Exception
            <DebuggerHidden()> _
            Get
                Return _error
            End Get
            <DebuggerHidden()> _
            Set(ByVal value As Exception)
                _error = value
            End Set
        End Property

        Protected Property Scanner() As Parser.Scanner
            <DebuggerHidden()> _
            Get
                Return _scanner
            End Get
            <DebuggerHidden()> _
            Set(ByVal value As Parser.Scanner)
                _scanner = value
            End Set
        End Property

#End Region

        <DebuggerHidden()> _
        Private Shared Function ExpressionFromFormat( _
            ByVal expression As String, _
            ByVal start As Integer, _
            ByVal format As ExpressionFormat _
        ) As Expression
            Select Case format
                Case ExpressionFormat.Infix
                    Return New InfixExpression(expression, start)
                Case ExpressionFormat.Prefix
                    Return New PrefixExpression(expression, start)
                Case ExpressionFormat.Postfix
                    Return New PostfixExpression(expression, start)
            End Select

            Return Nothing
        End Function

#End Region

    End Class

End Namespace