Namespace Utility.Functional

    Public Delegate Function UnaryOperation(Of I, O)(ByVal x As I) As O
    Public Delegate Function BinaryOperation(Of A, B, C)(ByVal x As A, ByVal y As B) As C

    ''' <summary>
    ''' This module contains a collection of "higher order" list functions similar to some found in the
    ''' Haskell Prelude. The functions facilitate application of binary operators to a whole list of values.
    ''' </summary>
    ''' <remarks>
    ''' Some of the functions were used in the course of development but are, at the moment, no longer used.
    ''' I do not remove them in case they are needed later.
    ''' </remarks>
    Public Module FirstClassFunctions

        Private Function [AndAlso](ByVal a As Boolean, ByVal b As Boolean) As Boolean
            Return a AndAlso b
        End Function

        Private Function [OrElse](ByVal a As Boolean, ByVal b As Boolean) As Boolean
            Return a OrElse b
        End Function

        Public Function [And](ByVal l As List(Of Boolean)) As Boolean
            Return FoldRight(l, True, AddressOf [AndAlso])
        End Function

        Public Function [Or](ByVal l As List(Of Boolean)) As Boolean
            Return FoldRight(l, False, AddressOf [OrElse])
        End Function

        Public Function Any(Of T)(ByVal l As List(Of T), ByVal p As Predicate(Of T)) As Boolean
            Return [Or](Map(Of T, Boolean)(l, p))
        End Function

        Public Function All(Of T)(ByVal l As List(Of T), ByVal p As Predicate(Of T)) As Boolean
            Return [And](Map(Of T, Boolean)(l, p))
        End Function

        Public Function FoldLeft(Of A, B)( _
            ByVal l As List(Of B), _
            ByVal v As A, _
            ByVal op As BinaryOperation(Of A, B, A) _
        ) As A
            Dim ret As A = v

            For i As Integer = 0 To l.Count - 1
                ret = op(ret, l(i))
            Next

            Return ret
        End Function

        Public Function FoldRight(Of A, B)( _
            ByVal l As List(Of A), _
            ByVal v As B, _
            ByVal op As BinaryOperation(Of A, B, B) _
        ) As B
            Dim ret As B = v

            For i As Integer = l.Count - 1 To 0 Step -1
                ret = op(l(i), ret)
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Executes an operation for each member of a list.
        ''' This is performed on a copy, leaving the original list untouched.
        ''' </summary>
        ''' <typeparam name="I">Member type of the input list.</typeparam>
        ''' <typeparam name="O">Member type of the resulting list.</typeparam>
        ''' <param name="a">The list.</param>
        ''' <param name="op">The unary operation.</param>
        ''' <returns>Returns the modified list.</returns>
        Public Function Map(Of I, O)( _
            ByVal a As List(Of I), _
            ByVal op As UnaryOperation(Of I, O) _
        ) As List(Of O)
            Dim ret As New List(Of O)()

            For Each x As I In a
                ret.Add(op(x))
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Executes an operation for each member of a list.
        ''' This is performed on a copy, leaving the original list untouched.
        ''' </summary>
        ''' <typeparam name="I">Member type of the input list.</typeparam>
        ''' <typeparam name="O">Member type of the resulting list.</typeparam>
        ''' <param name="a">The list.</param>
        ''' <param name="op">The unary operation.</param>
        ''' <returns>Returns the modified list.</returns>
        Public Function Map(Of I, O)( _
            ByVal a As List(Of I), _
            ByVal op As [Delegate] _
        ) As List(Of O)
            Dim ret As New List(Of O)()

            For Each x As I In a
                ret.Add(DirectCast(op.DynamicInvoke(x), O))
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Maps a value via an operation to each member of a list.
        ''' In this binary operation, the value is the second parameter.
        ''' This is performed on a copy, leaving the original list untouched.
        ''' </summary>
        ''' <typeparam name="A">Member type of the input list.</typeparam>
        ''' <typeparam name="B">Type of the value.</typeparam>
        ''' <typeparam name="C">Member type of the resulting list.</typeparam>
        ''' <param name="l">The list.</param>
        ''' <param name="v">The value.</param>
        ''' <param name="op">The binary operation.</param>
        ''' <returns>Returns the modified list.</returns>
        Public Function MapValueRight(Of A, B, C)( _
            ByVal l As List(Of A), _
            ByVal v As B, _
            ByVal op As BinaryOperation(Of A, B, C) _
        ) As List(Of C)
            Dim ret As New List(Of C)()

            For Each x As A In l
                ret.Add(op(x, v))
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Maps a value via an operation to each member of a list.
        ''' In this binary operation, the value is the first parameter.
        ''' This is performed on a copy, leaving the original list untouched.
        ''' </summary>
        ''' <typeparam name="A">Member type of the input list.</typeparam>
        ''' <typeparam name="B">Type of the value.</typeparam>
        ''' <typeparam name="C">Member type of the resulting list.</typeparam>
        ''' <param name="l">The list.</param>
        ''' <param name="v">The value.</param>
        ''' <param name="op">The binary operation.</param>
        ''' <returns>Returns the modified list.</returns>
        Public Function MapValueLeft(Of A, B, C)( _
            ByVal l As List(Of A), _
            ByVal v As B, _
            ByVal op As BinaryOperation(Of B, A, C) _
        ) As List(Of C)
            Dim ret As New List(Of C)()

            For Each x As A In l
                ret.Add(op(v, x))
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Combines two lists by applying an operation to the members pairwisely.
        ''' In this operation, both lists are seen as closed cycles so when one list is shorter than
        ''' the other, its members are continuously repeated until its length matches the other list.
        ''' </summary>
        ''' <typeparam name="A">Member type of the first list.</typeparam>
        ''' <typeparam name="B">Member type of the second list.</typeparam>
        ''' <typeparam name="C">Member type of the resulting list.</typeparam>
        ''' <param name="x">The first list.</param>
        ''' <param name="y">The second list.</param>
        ''' <param name="op">The binary operation.</param>
        ''' <returns>Returns the resulting list.</returns>
        ''' <remarks>
        ''' The semantics of this function differs from that found in the Haskell Prelude with the same
        ''' name, as its list parameters are seen as circles rather than enforcing an equal length for them.
        ''' </remarks>
        Public Function ZipWith(Of A, B, C)( _
            ByVal x As List(Of A), _
            ByVal y As List(Of B), _
            ByVal op As BinaryOperation(Of A, B, C) _
        ) As List(Of C)
            Dim ret As New List(Of C)()
            Dim i1 As Integer = 0
            Dim i2 As Integer = 0
            Dim len As Integer

            If x.Count > y.Count Then len = x.Count Else len = y.Count

            For i As Integer = 1 To len
                ret.Add(op(x(i1), y(i2)))

                i1 = (i1 + 1) Mod x.Count
                i2 = (i2 + 1) Mod y.Count
            Next

            Return ret
        End Function

        ''' <summary>
        ''' Combines n lists by applying an n-ary operation to the members pairwisely.
        ''' In this operation, all lists are seen as closed cycles so when one list is shorter than
        ''' the others, its members are continuously repeated until its length matches the other lists.
        ''' </summary>
        ''' <typeparam name="I">Member type of the input lists.</typeparam>
        ''' <typeparam name="O">Member type of the resulting list.</typeparam>
        ''' <param name="ls">The input lists.</param>
        ''' <param name="op">The n-ary operation.</param>
        ''' <returns>Returns the resulting list.</returns>
        Public Function ZipWithN(Of I, O)( _
            ByVal ls As List(Of I)(), _
            ByVal op As [Delegate] _
        ) As List(Of O)
            Dim ret As New List(Of O)()
            Dim idxs(ls.Length - 1) As Integer ' Current incides of all lists
            Dim len As Integer = 0

            ' Determine the maximum length.

            For Each l As List(Of I) In ls
                If l.Count > len Then len = l.Count
            Next

            ' Iterate through the values

            For idx As Integer = 1 To len
                Dim params(ls.Length - 1) As I

                For j As Integer = 0 To ls.Length - 1
                    params(j) = ls(j).Item(idxs(j))
                    idxs(j) = (idxs(j) + 1) Mod ls(j).Count
                Next

                ret.Add(DirectCast(op.DynamicInvoke(params), O))
            Next

            Return ret
        End Function

    End Module

End Namespace